#
# install.py - SNAKE routines for setting up and performing an installation
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import os, gzip, tempfile
import shutil
import snake.uri
import snake.util
from snake.translate import _
from urlgrabber import grabber

# FIXME this needs to support various arches
def grubby(kernel, initrd, args='', title='snake', arch=None, progresscb=None):
    '''Uses grubby to modify the bootloader config to make the given kernel
    and initrd the default bootloader choice. Uses 'snake' for the title if one
    is not given.'''

    if progresscb:
        progresscb.start(text=_("Updating bootloader"), size=2)
        progresscb.update(0)

    # FIXME check to see if 'snake' already exists and add a suffix if needed
    cmd="""/sbin/grubby --make-default --title='%s' --add-kernel=%s --initrd=%s --args='%s'""" % (title,kernel,initrd,args)
    rc = os.system(cmd)
    if progresscb: progresscb.update(1)

    if arch.startswith("ppc"):
        rc = os.system("/sbin/ybin")
    elif arch.startswith("s390"):
        rc = os.system("/sbin/zipl")

    if progresscb:
        progresscb.update(2)
        progresscb.end(2)

    return rc

def __file_write(inf,outf,blocksize=4096):
    '''Write the contents of inf to outf'''
    bytes = 0
    while True:
        data = inf.read(blocksize)
        if data:
            outf.write(data)
            bytes = bytes + len(data)
        else:
            break
    inf.close()
    outf.close()
    return bytes

def add_file_to_initrd(addfile,initrd,outfile=None,progresscb=None):
    '''Adds the given file to the given initrd. This encodes the relative
    path of the added file into the initrd, so if you want to add a file foo
    in usr/share on the initrd, it needs to be in usr/share/foo locally.'''

    if progresscb: progresscb.start(text=_("Building initrd"), size=6)
    if progresscb: progresscb.update(0)

    # Create temporary file
    (fd,tmpinitrd) = tempfile.mkstemp('.cpio')
    os.close(fd)
    # uncompress initrd to tmpinitrd
    b = __file_write(gzip.open(initrd),open(tmpinitrd,'wb'))
    if progresscb: progresscb.update(1)

    # perhaps this is a gzip'd ext2 file system?
    # NOTE pyparted works as well ... but is not included in a @base install
    tmpext2fs = tempfile.mkdtemp('.loop')
    isMounted = 0 == os.system("mount -o loop -t ext2 %s %s 2>/dev/null" % (tmpinitrd, tmpext2fs))
    if isMounted:
        add_command = "cat %(addfile)s > %(initrd)s/%(addfile)s"
    else:
        add_command = "pax -w -a -f %(initrd)s %(addfile)s"
    if progresscb: progresscb.update(2)

    # add file to tmpinitrd
    r = os.system(add_command % {"initrd": isMounted and tmpext2fs or tmpinitrd, "addfile": addfile})
    if r != 0: # IT'S ALL GONE WRONG!!
        os.unlink(tmpinitrd)
        return False
    if progresscb: progresscb.update(3)

    # if needed, umount ext2 initrd
    if isMounted and os.path.isdir(tmpext2fs):
        os.system("umount %s" % tmpext2fs)
    if progresscb: progresscb.update(4)

    # compress tmpinitrd to outfile
    if not outfile:
        # Nothing specified - overwrite the input file
        outfile = initrd
    b = __file_write(open(tmpinitrd),gzip.open(outfile,'wb'))
    if progresscb: progresscb.update(5)

    # clean up
    if os.path.isdir(tmpext2fs):
        os.rmdir(tmpext2fs)
    os.unlink(tmpinitrd)

    if progresscb: progresscb.update(6)
    if progresscb: progresscb.end(6)

    return True

def fetch_images(images, fetcher, progresscb):
    '''Fetches boot images from the given URI for the given arch, then updates
    the bootloader config to boot into the installer.
    Accepted arguments:
      images (list): list of dictionaries.  Each dictionary should contain a
      file type and relative path
      fetcher (instance): an instance of snake.uri.fetcher to handle
      downloading files
      progresscb (instance): used for progress notification
    '''

    fnames = list()

    # .discinfo trees may provide a list of 'kernel' and 'initrd' images
    for image_type in ['kernel', 'initrd']:
        # coerce as a list
        if not isinstance(images[image_type], list):
            images[image_type] = [images[image_type]]

        # loop through list of paths
        for image_path in images[image_type]:
            try:
                local_path = fetcher.saveFile(image_path,progresscb)
            except IOError, e:
                local_path = None
                continue # try the next image
            else:
                fnames.append(local_path)
                break # once we get a valid image_path stop looking

    if len(fnames) != 2:
        raise ValueError(_("Failed to fetch boot images"))

    return fnames

def fetch_and_prep(uri,**kwargs):
    '''Fetches boot images from the given URI for the given arch,
    then updates the bootloader config to boot into the installer.
    Accepted kwargs:
    args=string: append given args to the boot commandline.
    arch=string: get images for the given arch (e.g. 'xen') instead of the
                 default tree arch.
    progress_callback=instance: call the given function with various messages
                                used to indicate progress.
    ksdata=string: the given string is kickstart data, to be embedded in the
                   initrd and used during installation.'''

    # Establish a progress callback
    progresscb = kwargs.get('progress_callback', None)

    # Create a fetcher to process the URI
    fetcher = snake.uri._fetcherForURI(uri, "/boot")
    fetcher.prepareLocation(progresscb)

    try:
        # Did the caller provide a snake.tree.Tree() object?
        if kwargs.has_key('tree'):
            tree = kwargs.get('tree')

        # Otherwise, we need to get a Tree() object
        else:
            tree = snake.uri.uri_to_tree(uri, fetcher)

        # Determine the architecture
        arch = kwargs.get('arch', tree.arch)

        # Determine where to place downloaded files
        if arch == "ia64":
            fetcher.scratchdir = "/boot/efi/efi/redhat"

        # Find boot images
        if not tree.images.has_key(arch):
            raise ValueError(_("No bootable images found."))

        (kernel, initrd) = fetch_images(tree.images[arch], fetcher, progresscb)

        # FIXME: get default args in a smarter way
        # consider also 'ipv6=auto'
        args = kwargs.get("args","ksdevice=link ip=dhcp noipv6")

        # Are we doing a kickstart install?
        if kwargs.has_key('ksdata') and kwargs.get('ksdata'):
            tmpdir = tempfile.mkdtemp()
            cwd = os.getcwd()
            os.chdir(tmpdir)
            outf = open('ks.cfg','w')
            outf.write(kwargs['ksdata'])
            outf.close()

            # where to put the kickstart file?
            ksmethod = kwargs.get("ksmethod","initrd")

            # On remote volume
            if ksmethod in ("http","ftp","nfs"):
                '''no action required.  The ks url should already be on the
                boot args'''

            # On local disk - defaults to /boot partition
            elif ksmethod.startswith("hd"):
                # FIXME - What about ia64 efi path?
                # FIXME - Should be a common variable ... needed in other components
                ksfile = os.path.join("/boot","ks.cfg")
                try:
                    (ksdev,kspath) = snake.util.get_file_device_path(ksfile)
                    shutil.copyfile(os.path.join(tmpdir,"ks.cfg"), ksfile)
                    # remove "/dev/" from the device string per anaconda documentation
                    if ksdev[:5] == "/dev/": ksdev = ksdev[5:]
                    args = args + " ks=hd:%s:%s" % (ksdev,kspath)
                except Exception, e:
                    raise ValueError(_("Error copying kickstart data to %s - %s" % (ksfile, e)))

            # Inside initrd
            elif ksmethod == "initrd":
                if not add_file_to_initrd('ks.cfg', initrd, progresscb=progresscb):
                    raise ValueError(_("Error embedding kickstart data into initrd"))
                args = args + " ks=file:/ks.cfg"

            # Error - unknown method requested
            else:
                raise ValueError(_("Unrecognized kickstart delivery method: %s" % ksmethod))
        else:
            args = args + " method=%s" % uri

        snake.install.grubby(kernel, initrd, args, title="install %s" % str(tree), arch=arch, progresscb=progresscb)

    finally:
        fetcher.cleanupLocation()
